import os
import sys

import pyhectiqlab
from pyhectiqlab.metrics import MetricsManager
from pyhectiqlab.settings import app_url
from pyhectiqlab.events_manager import EventsManager
from pyhectiqlab.stream import LogHandler
from pyhectiqlab.buffer import Buffer
from pyhectiqlab.watermark import Watermark
from pyhectiqlab.artifacts import SharedArtifactsManager
from pyhectiqlab.notebooks import get_notebook_file
from pyhectiqlab.utils import list_all_files_in_dir

from pyhectiqlab.mlmodels import download_mlmodel as ops_download_mlmodel
from pyhectiqlab.mlmodels import upload_mlmodel as ops_upload_mlmodel
from pyhectiqlab.datasets import download_dataset as ops_download_dataset
from pyhectiqlab.datasets import upload_dataset as ops_upload_dataset
from pyhectiqlab.mlmodels import mlmodel_info_at_path

from functools import partial, wraps
from typing import Optional
from sys import getsizeof
import packaging 
import tempfile
import logging
import shutil

class Run():
    
    def __init__(self, name: str, project: str = None, should_create_project: bool = False, dry: bool = False):
        """
        name: Run name. If a run exists with this name, you'll be attached to the existing run. Otherwise a 
            new run is created.
        project: The project name. Leave None if the user with the API key has access to a single project. 
            Otherwise, you must specify the project name.
        should_create_project: bool. If true, creates a project d if if does not exists (Admin only).
        dry: bool: If true, the run is in a dry mode (no writting).
        
        """
        self._mock = dry
        self.events_manager = EventsManager(self._mock)
        # Check the python version
        self.events_manager.compare_python_version()  

        if dry:
            print("Dry mode activated.")
            self.run_view = {"id": "null"}
            self._mock = True
            self.events_manager = EventsManager(self._mock)
            self._read_only = False
        elif self.events_manager.is_logged()==False:
            print("User not authentificated. Use `pyhectiqlab.login()` for a web prompt or `hectiqlab login` in command line.")
            self._failed_init()
        else:
            # try:
            if project is None:
                projects = self.events_manager.add_event("get_all_projects", args=(), auth=True, async_method=False)
                if len(projects["result"])==1:
                    project = projects["result"][0]["name"]
                    print(f"Connecting to project {project}.")
                    self.project_view = projects["result"][0]
                else:
                    print("User has access to multiple projects. Switching to dry mode. \n--------\nFor active mode, please use a project:")
                    for el in projects["result"]:
                        print(f"{el['name']}")
                    self._failed_init()
                    pass
            else:
                self.project_view = self.events_manager.add_event("create_project", args=(project, should_create_project), auth=True, async_method=False)
                if self.project_view is None:
                    print("Project does not exists. Dry mode activated. \n--------\nUse should_create_project=True to create a project.")
                    self._failed_init()
                    pass
                if self.project_view['status_code']==401:
                    self._failed_init()
                    pass
            if project:
                self.run_view = self.events_manager.add_event("create_run", args=(name, project), auth=True, async_method=False)
                try:
                    if self.run_view["is_new"]==False:
                        print(f"Attaching to an existing run (ID :{self.run_view['p_id']}).")
                    if self.run_view['readonly']:
                        print(f"Run is read only.")
                    self._read_only = self.run_view['readonly']
                except:
                    print("An error occured while creating the run. Switching to dry mode.")
                    self._failed_init()

                if "id" not in self.run_view:
                    print("An error occured while creating the run. Switching to dry mode.")
                    self._failed_init()

        self.name = name
        self.metrics_manager = MetricsManager(run_id=self.run_view["id"], push_method=self._push_metrics)
        self.metrics_manager.set_aggr('mean')
        self.metrics_manager.update_cache_settings(max_cache_timeout=20,
                                                    min_cache_flush_delay=5, 
                                                    max_cache_length=25)
        self.log_recorder = None
        self.is_recording_logs = False
        self.logs_buffer = Buffer(push_method=self._push_logs)
        return

    def write_method(f):
        @wraps(f)
        def wrapper(self, *args, **kwargs):
            if self.read_only:
                print('Run is read only. No actions.')
                return
            return f(self, *args, **kwargs)
        return wrapper

    def beta(f):
        @wraps(f)
        def wrapper(self, *args, **kwargs):
            print('This method is still in development and can be unstable.')
            return f(self, *args, **kwargs)
        return wrapper

    def will_be_depreciated(suggested_method):
        def _method(f):
            @wraps(f)
            def wrapper(self, *args, **kwargs):
                print(f'This method will be removed in a future release. Use {suggested_method} instead.')
                return f(self, *args, **kwargs)
            return wrapper
        return _method

    def set_metrics_cache_settings(self, aggr: str = 'mean', max_cache_timeout=None, max_cache_length=None, min_cache_flush_delay=None):
        """
        Update the settings of the metrics manager.

        max_cache_timeout [int]: Time interval for flushing the cache (seconds)  (default is 20)
        max_cache_length [int]: Max number of elements before flushing the cache (default is 25)
        min_cache_flush_delay [float]: Minimum delay between to successive data send (for the same key). 
                If the number of matrix is added faster than this delay, the cache will increase to
                respect the delay. (default is 5 seconds)
        aggr [str]: The aggregate method. One of ['none', 'sum', 'max', 'mean']  (default is mean)
        """
        self.metrics_manager.set_aggr(aggr)
        self.metrics_manager.update_cache_settings(max_cache_timeout=max_cache_timeout, 
                                                    max_cache_length=max_cache_length,
                                                    min_cache_flush_delay=min_cache_flush_delay)

    def _failed_init(self):
        self.run_view = {"id": "null"}
        self._mock = True
        self.events_manager = EventsManager(self._mock)
        self._read_only = True

    @write_method
    def add_artifacts(self, filepath: str, step: int = None):
        """Log a file as an artifacts. If a file already exists with the name, it will be overwritten.

        filepath [str] : Path to the file.
        step [int]: The optional step stamp of the artifacts.
        """
        filename = os.path.basename(filepath)
        num_bytes = os.path.getsize(filepath)
        content_bytes = open(filepath, "rb")
        args = (self._id, filename, content_bytes, num_bytes, step)
        self.events_manager.add_event("add_file", args)

    @write_method
    @beta
    def add_mlmodel_usage(self, mlmodel_name: str, version: str = None):
        """
        """
        args = (self._id, mlmodel_name, version, self.project_view['id'])
        self.events_manager.add_event("log_mlmodel", args)

    @write_method
    @beta
    def add_mlmodel_usage_from_dirpath(self, dirpath:str):
        """Log that this run is using a certain model. 
        """
        meta = mlmodel_info_at_path(dirpath)
        assert meta is not None, f'Could not find the mlmodel at {dirpath}'

        print(f"{meta.get('name')} - {meta.get('version')}")
        args = (self._id, meta.get('name'), meta.get('version'), self.project_view['id'])
        self.events_manager.add_event("log_mlmodel", args)
        return
        
    @write_method        
    def add_mlmodel(self, source_path: str, name: str, version: str = None, short_description: str = None,
                push_dir: bool = False):
        """Add a mlmodel.

        source_path: Path to the directory/file of the mlmodel
        name: Name of the model without spaces and backslash (e.g., 'text-model')
        version: Version in format '{major}.{minor}.{micro}' (e.g., 1.2.0). If None, 
            the version 1.0.0 is assigned or an increment of minor of the latest version of the
            model with this name (e.g. 1.3.3 -> 1.4.0)
        short_description: Short description.
        push_dir: Must be set to True if the source_path is a directory
        """
        ops_upload_mlmodel(source_path=source_path, 
            mlmodel_name=name,
             run_id=self._id, 
             version=version,
             short_description=short_description,
             push_dir=push_dir)

    def download_mlmodel(self, mlmodel_name: str, version: str = None, save_path: str = './', overwrite:bool = False):
        """Download an existing mlmodel from the run's project.

        source_path: Path to the directory/file of the mlmodel
        mlmodel_name: Name of the model without spaces and backslash (e.g., 'text-model')
        version: Specific version of the model. If None, the latest version is fetched.
        save_path: Path to where the model is saved.
        overwrite: Set to True to overwrite in save_path.
        """
        dirpath = ops_download_mlmodel(mlmodel_name=mlmodel_name, 
                        project_id=self.project_view["id"], 
                        version=version, 
                        save_path=save_path, 
                        overwrite=overwrite)
        if dirpath is not None:
            print(dirpath) 
        return dirpath
        
    @write_method
    def add_dataset(self, source_path: str, name: str, version: str = None, short_description: str = None,
                push_dir: bool = False):
        """Add a dataset.

        source_path: Path to the directory/file of the dataset
        name: Name of the dataset without spaces and backslash (e.g., 'text-dataset')
        version: Version in format '{major}.{minor}.{micro}' (e.g., 1.2.0). If None, 
            the version 1.0.0 is assigned or an increment of minor of the latest version of the
            dataset with this name (e.g. 1.3.3 -> 1.4.0)
        short_description: Short description.
        push_dir: Must be set to True if the source_path is a directory
        """
        ops_upload_dataset(source_path=source_path, 
            dataset_name=name,
             run_id=self._id, 
             version=version,
             short_description=short_description,
             push_dir=push_dir)

    def download_dataset(self, dataset_name: str, version: str = None, save_path: str = './', overwrite:bool = False):
        """Download an existing dataset from the run's project.

        source_path: Path to the directory/file of the dataset
        mlmodel_name: Name of the dataset without spaces and backslash (e.g., 'text-dataset')
        version: Specific version of the dataset. If None, the latest version is fetched.
        save_path: Path to where the dataset is saved.
        overwrite: Set to True to overwrite in save_path.
        """
        dirpath = ops_download_dataset(dataset_name=dataset_name, 
                        project_id=self.project_view["id"], 
                        version=version, 
                        save_path=save_path, 
                        overwrite=overwrite)
        if dirpath is not None:
            print(dirpath)        
        return dirpath
    
    @write_method
    def add_config(self, config: 'Config', prefix=""):
        """Log a config object (pyhectiqlab.Config). Use prefix 
        to log a config in a nested path.
        """
        config.push_as_meta(self, prefix=os.path.join("config/", prefix))

    @write_method
    @will_be_depreciated('add_tf_model_as_artifact')
    def add_tf_model(self, model: 'tf.keras.Model', filename: 'str', step: int = None):
        """Add a tensorflow model as an artifact.
        """
        self.add_tf_model_as_artifact(model=model, filename=filename, step=step)

    @write_method
    def add_tf_model_as_artifact(self, model: 'tf.keras.Model', filename: str, step: int = None):
        """Add a tensorflow model as an artifact.
        """
        save_path = tempfile.mkdtemp()
        model.save_weights(f"{save_path}/{filename}")
        p = tempfile.mkdtemp()
        shutil.make_archive(f"{p}/{filename}", 'zip', save_path)
        self.add_artifacts(f"{p}/{filename}.zip", step=step)

    @write_method
    def add_directory_as_zip_artifact(self, dirpath: str, step: int = None):
        p = tempfile.mkdtemp()
        filename = os.path.basename(dirpath)
        shutil.make_archive(f"{p}/{filename}", 'zip', root_dir=dirpath)
        self.add_artifacts(f"{p}/{filename}.zip", step=step)

    @write_method
    def add_shared_artifact(self, filename: str):
        """Push an artifact to be shared accross the project.
        """
        m = SharedArtifactsManager()
        m.push_artifact(filepath=filename, project_id=self.project_view["id"])

    def list_shared_artifact(self):
        """Returns a list of the shared artifacts for this project.
        """
        m = SharedArtifactsManager()
        return m.list_artifacts(project_id=self.project_view["id"])

    @write_method
    def add_metrics(self, key: 'str', value: float, step: int):
        """Add a metrics. The metrics is saved even if there already exists a value
        for the same key and step. 

        key [str]: Key used for the metrics. The format with backlash such as `metrics/train`
            will be recognized to organize the metrics.
        value [float]. If tensor, is casted to float.
        step [int]: Training step. Casted to integer.
        """
        self.metrics_manager.add(key, float(value), int(step))

    @write_method
    def add_tag(self, name: str, description: str = None, color: str = None):
        """Add tag to a run. If the tag does not already exists by this name,
        a new tag is created with the given description and color.
        """
        args = (self._id, name, description, color)
        self.events_manager.add_event("add_tag", args)

    @write_method
    def add_package_versions(self, packages: Optional[dict] =None, with_sys: bool = True, with_python: bool = True, with_git: bool = True):
        """Save the version of the imported packages.

        Use:
        add_package_versions(globals())
        """
        water = Watermark()
        versions = {}
        if with_python:
            versions["python"] = water.get_pyversions()
        if with_sys:
            versions["system"] = water.get_sysinfo()
        if packages:
            versions["packages"] = water.get_all_import_versions(packages)
        if with_git:
            git_info = water.get_git_info()
            if git_info:
                versions["git"] = water.get_git_info()

        args = (self._id, versions)
        self.events_manager.add_event("push_package_versions", args)

    @write_method
    def add_package_repo_state(self, package: str = None):
        """Save the branch/commit/origin of the git repo in which
        a package is located.

        Example:
            self.add_package_repo_state('pyhectiqlab')
        """
        water = Watermark()
        result = water.get_repo_info(package)
        data = {package: result}
        args = (self._id, data)
        self.events_manager.add_event("push_git_package_state", args)

    @write_method
    def add_current_notebook(self, stamp: str = 'datetime'):
        """If the run is executed in a jupyter notebook, this
        will add your current notebook as an artifact.

        stamp [str]: Add a stamp on the notebook name.
                    One of 'datetime', 'date', None
        """
        filename = get_notebook_file(stamp)
        self.add_artifacts(filename)

    @write_method
    def failed(self):
        """Set the run to the failed status.
        """
        args = (self._id, "failed")
        self.events_manager.add_event("set_run_status", args)

    @write_method
    def _stopped(self):
        """Set the run to the stopped status.
        """
        args = (self._id, "stopped")
        self.events_manager.add_event("set_run_status", args)
        
    @write_method
    def completed(self):
        """Set the run to the completed with success status.
        """
        args = (self._id, "completed")
        self.events_manager.add_event("set_run_status", args)
        
    @write_method
    def pending(self):
        """Set the run to the pending status.
        """
        args = (self._id, "pending")
        self.events_manager.add_event("set_run_status", args)
        
    @write_method
    def running(self):
        """Set the run to the running status.
        """
        args = (self._id, "running")
        self.events_manager.add_event("set_run_status", args)
        
    @write_method
    def training(self):
        """Set the run to the training status.
        """
        args = (self._id, "training")
        self.events_manager.add_event("set_run_status", args)
        
    @write_method
    def add_meta(self, key, value):
        """Add or update a meta information in the format of
        key/value.
        """ 
        args = (self._id, key, value)
        self.events_manager.add_event("push_meta", args)
        
    @write_method
    def set_note(self, text):
        """Add or update a custom note field for the run.
        """
        args = (self._id, text)
        self.events_manager.add_event("set_note", args)
        
    @write_method
    def start_recording_logs(self, logger_name=None):
        """Start tailing the logs on a logger name.
        If logger_name is None, the default __name__ will be taken and
        you may use the returned logger.
        """
        self.is_recording_logs = True
        self.logs_buffer.start(key="logs")
        
        print("Start recording logs.")
        if (self.log_recorder is None) or self.logger_name!=logger_name:
            func = partial(self.logs_buffer.add, key="logs")
            self.log_recorder = LogHandler(func)
            console  = logging.StreamHandler()  
            console.setLevel(logging.DEBUG)

            name = logger_name
            if name is None:
                name = __name__
            self.logger_name = name

            logger = logging.getLogger(name)
            logger.setLevel(logging.DEBUG)
            logger.addHandler(self.log_recorder)

            # Handle all exceptions
            def handle_exception(exc_type, exc_value, exc_traceback):
                if issubclass(exc_type, KeyboardInterrupt):
                    self._stopped()
                else:
                    self.add_tag(name="failed")
                    self.failed()
                    
                logger.error("Exception", exc_info=(exc_type, exc_value, exc_traceback))
                # Flush cache before leaving
                self.logs_buffer.flush_cache()
                self.metrics_manager.flush_cache()
                sys.exit(0)                
                
            sys.excepthook = handle_exception

            return logger
        else:
            return logging.getLogger(self.logger_name)
        
    @write_method
    def stop_recording_logs(self):
        """Stop recording the logs from the logger. 
        """
        self.is_recording_logs = False
        self.logs_buffer.stop(key="logs")
        
    @write_method
    def _push_logs(self, key, msg):
        if key=="logs":
            if self.is_recording_logs:
                args = (self._id, "".join(msg))
                self.events_manager.add_event("append_logs", args)
        
    @write_method
    def _push_metrics(self, key, values):
        args = (self._id, key, values)
        self.events_manager.add_event("push_metrics", args)

    @property
    def _id(self):
        return self.run_view["id"]

    @property
    def read_only(self):
        return self._read_only

    @property
    def dry_mode(self):
        return self._mock

    @property
    def action_mode(self):
        if self.dry_mode:
            return 'dry'
        if self.read_only:
            return 'read-only'
        return 'read-write'
    
    @property
    def author(self):
        author = self.run_view.get('author')
        if author is None:
            return 'Undefined'
        return author.get('firstname') + ' ' + author.get('lastname')
        
    def __str__(self):
        # print(self.run_view)
        pad = 9
        path = f"/{self.project_view['id']}/runs/{self._id}"
        return f"<Run {self.run_view['p_id']}>"\
                f"\n{'project'.ljust(pad)}: {self.project_view['name']}"\
                f"\n{'author'.ljust(pad)}: {self.author}"\
                f"\n{'mode'.ljust(pad)}: {self.action_mode}"\
                f"\n{'url'.ljust(pad)}: {app_url+path} "

    def __repr__(self):
        return self.__str__()
    


import os
from pyhectiqlab.utils import get_single_user_project, list_all_files_in_dir, load_event_manager

import requests
from tqdm import tqdm
import logging
import json
import datetime as dt
import packaging

logger = logging.getLogger('pyhectiqlab')
logger.setLevel(logging.WARNING)

def mlmodel_info_at_path(dirpath: str):
    try:
        with open(os.path.join(dirpath, '.hlab_meta.json'), "r") as f:
            data = json.load(f)
        return data
    except:
        return None

def upload_mlmodel(source_path: str, mlmodel_name: str, run_id: str = None, version: str = None,
    short_description: str = None, push_dir:bool = False):
    assert ' ' not in mlmodel_name, f"The name cannot contain spaces (use `{mlmodel_name.replace(' ','-')}`)"
    assert '/' not in mlmodel_name, f"The name cannot contain / (use `{mlmodel_name.replace('/','-')}`)"

    manager = load_event_manager()

    if version is not None:
        v = packaging.version.parse(version)
        assert isinstance(v, packaging.version.Version), 'Cannot use legacy version names. Use a format major.minor.micro.'
        updated_version = f'{v.major}.{v.minor}.{v.micro}'
        assert updated_version==version, f'Please change version to {updated_version}.'

    if os.path.isdir(source_path):
        print('Model is a directory. Listing all files...')
        paths = list_all_files_in_dir(source_path)
        filenames = [os.path.relpath(p, start=source_path) for p in paths]
        num_bytes = [os.path.getsize(p) for p in paths]
        full_paths = {f: os.path.abspath(p) for f,p in zip(filenames, paths)}
        print(f'Found {len(filenames)} files')

        assert push_dir, 'Set push_dir=True to push a model directory.'

    else:
        filenames = [os.path.basename(source_path)]
        full_paths = {os.path.basename(source_path): os.path.abspath(source_path) }
        num_bytes = os.path.getsize(source_path)
    args = (run_id, filenames, full_paths, num_bytes, mlmodel_name, short_description, version)
    manager.add_event("add_mlmodel", args, async_method=False)

def download_mlmodel(mlmodel_name: str, project_id: str = None, version: str = None, 
    save_path: str = './', overwrite: bool = False):

    if project_id is None:
        project = get_single_user_project()
        if project is None:
            logger.error('Cannot download mlmodel without a specific project.')
            return
        project_id = project.get('id')

    manager = load_event_manager()

    # Fetch the download info
    res = manager.add_event('get_mlmodel_download_info', 
            args=(mlmodel_name, project_id, version), 
            async_method=False)

    assert res.get('status_code')==200, res.get('detail')
    downloaded_version = res.get('meta').get('version')

    dirpath = os.path.join(save_path, mlmodel_name+'-'+downloaded_version)

    if version is None:
        logger.info(f'Fetching version {downloaded_version}')

    if os.path.isdir(dirpath) and overwrite==False:
        logger.error(f'Model is already downloaded at {dirpath}. Delete this folder or set overwrite==True to download again.')
        return

    if os.path.isdir(dirpath)==False:
        os.makedirs(dirpath)

    meta = res.get('meta')
    text = f"{meta.get('name')}\nVersion: {meta.get('version')}\nProject: {project_id}\n"
    text += f"UUID: {meta.get('uuid')}\nDescription: {meta.get('description')}\n"
    text += f"Downloaded on: {dt.datetime.now()}"
    text += '\n--------------------------\n'
    if meta.get('full_description'):
        text += meta.get('full_description')
    else:
        text += 'No description.'
    with open(os.path.join(dirpath, 'HLab_README.md'), "w") as f:
        f.write(text)

    with open(os.path.join(dirpath, '.hlab_meta.json'), "w") as f:
        json.dump(meta, f)

    # Download the files
    files = res.get('files')
    logger.info(f'Downloading {len(files)} files')
    for file_meta in files:
        url = file_meta['url']
        name = file_meta['name']
        if os.path.dirname(name)!='':
            subdirpath = os.path.join(dirpath, os.path.dirname(name))
            if os.path.isdir(subdirpath) == False:
                os.makedirs(subdirpath)
        logger.info(name)
        response = requests.get(url, stream=True)
        total_size_in_bytes= int(response.headers.get('content-length', 0))
        block_size = 1024
        progress_bar = tqdm(total=total_size_in_bytes, unit='iB', unit_scale=True)

        path = os.path.join(dirpath, name)
        with open(path, 'wb') as file:
            for data in response.iter_content(block_size):
                progress_bar.update(len(data))
                file.write(data)
        progress_bar.close()
        if total_size_in_bytes != 0 and progress_bar.n != total_size_in_bytes:
            logger.error("ERROR, something went wrong when downloading the file.")
        
        logger.info(f'Content saved at {path}')
    return dirpath



import os
import pickle as pkl

class Config():
    
    def __init__(self, **kwargs):
        for key in kwargs:
            if key=="dict":
                raise KeyError("Key dict is a protected attribute.")
            if key=="keys":
                raise KeyError("Key keys is a protected attribute.")
            setattr(self, key, kwargs[key])
    
    def save(self, path):
        pkl.dump(self.__dict__, open(path, "wb"))
    
    @property
    def default_run_dir(self):
        msg = "log_dir must be set."
        assert self.log_dir is not None, msg

        msg = "name must be set."
        assert self.name is not None, msg

        return os.path.join(self.log_dir, self.name)
    
    @property
    def default_model_path(self):
        return os.path.join(self.default_run_dir, "model.pt")
    
    @property
    def default_config_path(self):
        return os.path.join(self.default_run_dir, "config.pt")
    
    @classmethod
    def load(cls, path):
        return cls(**pkl.load(open(path, "rb")))
    
    @property
    def dict(self):
        return self.__dict__
    
    def __getattr__(self, name):
        if name in self.__dict__:
            return self.__dict__[name]
        return None
    
    def __getstate__(self): 
        return self.__dict__

    def __setstate__(self, d): 
        self.__dict__.update(d)

    def __repr__(self):
        return self.__str__()
    
    def keys(self):
        return self.__dict__.keys()
    
    def __getitem__(self, key):
        if key in self.__dict__:
            return self.__dict__[key]
        return None

    def push_as_meta(self, run, prefix=""):
        """Push the config as meta information to run.
        """
        for key, val in self.__dict__.items():
            if isinstance(val, Config):
                val.push_as_meta(run, prefix=prefix+f"{key}/")
            else:
                run.add_meta(f"{prefix}{key}", str(val))
                
    def __str__(self):
        s = "\n"
        for key, val in self.__dict__.items():
            if isinstance(val, Config):
                _s = "\n"+(" "+key+" ").ljust(20, '>').rjust(30, '>')
                _s += str(val)
                _s += ("").ljust(30, '<')+"\n"
                s += _s
            else:
                _s = key.ljust(20, ' ')+ ": "+ str(val)+"\n"
                s += _s
        return s


import os
import shutil
import tempfile
import datetime as dt

import json
import urllib.error
import urllib.request
from itertools import chain
from pathlib import Path, PurePath
from typing import Generator, Tuple, Union

import ipykernel
from jupyter_core.paths import jupyter_runtime_dir
from traitlets.config import MultipleInstanceError

"""
Code from:
https://github.com/msm1089/ipynbname

MIT License

Copyright (c) 2020 Mark McPherson

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

FILE_ERROR = "Can't identify the notebook {}."
CONN_ERROR = "Unable to access server;\n" \
           + "ipynbname requires either no security or token based security."


def _list_maybe_running_servers(runtime_dir=None) -> Generator[dict, None, None]:
    """ Iterate over the server info files of running notebook servers.
    """
    if runtime_dir is None:
        runtime_dir = jupyter_runtime_dir()
    runtime_dir = Path(runtime_dir)

    if runtime_dir.is_dir():
        for file_name in chain(
            runtime_dir.glob('nbserver-*.json'),  # jupyter notebook (or lab 2)
            runtime_dir.glob('jpserver-*.json'),  # jupyterlab 3
        ):
            yield json.loads(file_name.read_bytes())


def _get_kernel_id() -> str:
    """ Returns the kernel ID of the ipykernel.
    """
    connection_file = Path(ipykernel.get_connection_file()).stem
    kernel_id = connection_file.split('-', 1)[1]
    return kernel_id


def _get_sessions(srv):
    """ Given a server, returns sessions, or HTTPError if access is denied.
        NOTE: Works only when either there is no security or there is token
        based security. An HTTPError is raised if unable to connect to a 
        server.
    """
    try:
        qry_str = ""
        token = srv['token']
        if token:
            qry_str = f"?token={token}"
        url = f"{srv['url']}api/sessions{qry_str}"
        with urllib.request.urlopen(url) as req:
            return json.load(req)
    except Exception:
        raise urllib.error.HTTPError(CONN_ERROR)


def _find_nb_path() -> Union[Tuple[dict, PurePath], Tuple[None, None]]:
    try:
        kernel_id = _get_kernel_id()
    except (MultipleInstanceError, RuntimeError):
        return None, None  # Could not determine
    for srv in _list_maybe_running_servers():
        try:
            sessions = _get_sessions(srv)
            for sess in sessions:
                if sess['kernel']['id'] == kernel_id:
                    return srv, PurePath(sess['notebook']['path'])
        except Exception:
            pass  # There may be stale entries in the runtime directory
    return None, None


def name() -> str:
    """ Returns the short name of the notebook w/o the .ipynb extension,
        or raises a FileNotFoundError exception if it cannot be determined.
    """
    _, path = _find_nb_path()
    if path:
        return path.stem
    raise FileNotFoundError(FILE_ERROR.format('name'))


def get_notebook_path() -> Path:
    """ Returns the absolute path of the notebook,
        or raises a FileNotFoundError exception if it cannot be determined.
    """
    srv, path = _find_nb_path()
    if srv and path:
        root_dir = Path(srv.get('root_dir') or srv['notebook_dir'])
        return root_dir / path
    raise FileNotFoundError(FILE_ERROR.format('path'))


def get_notebook_file(stamp: str = 'datetime'):
    notebook_path = get_notebook_path()
    name, extension = os.path.splitext(notebook_path.parts[-1])
    if stamp == 'datetime':
        commit_version = dt.datetime.now().strftime('%Y-%m-%d_%H-%M')
        commit_name = f"{name}_v{commit_version}{extension}"
    elif stamp == 'date':
        commit_version = dt.datetime.now().strftime('%Y-%m-%d')
        commit_name = f"{name}_v{commit_version}{extension}"
    else:
        commit_name = f"{name}{extension}"
    commit_file = os.path.join(tempfile.gettempdir(), commit_name)
    shutil.copyfile(notebook_path, commit_file)
    return commit_file












from collections import defaultdict
from pyhectiqlab.timer import RepeatedTimer

import threading
import pyhectiqlab.ops as ops

class Buffer():

    def __init__(self, 
        max_cache_timeout=5,  # seconds
        max_cache_length=100, # Number of elements
        extra_args={},
        push_method=None):

        self.extra_args = extra_args
        self.cache = defaultdict(list)
        self.can_record = defaultdict(lambda: True)

        self.max_cache_length = max_cache_length
        self.push_method = push_method

        self.timer = RepeatedTimer(self.flush_cache, max_cache_timeout)
        self.timer.start()
        return

    def stop(self, key):
        self.can_record[key] = False

    def start(self, key):
        self.can_record[key] = True

    def add(self, value, key):
        if self.can_record[key]==False:
            return

        self.cache[key].append(value)
        if len(self.cache[key])>self.max_cache_length:
            self.push_data(key)

    def push_data(self, key):
        self.push_method(key, self.cache[key])
        self.cache[key] = []

    def __delete__(self):
        self.stop_flag.set()
        self.timer.cancel()
        self.flush_cache()

    def flush_cache(self):
        for key in self.cache:
            self.push_data(key)



import os
import toml
import datetime as dt
import pyhectiqlab.ops as ops

from pathlib import Path

class AuthProvider(object):
    """
    Class used to handle oAuth
    """
    def __init__(self, port=8082):
        self.port = str(port)
        self.profile = os.environ.get('HECTIQLAB_PROFILE') or 'default'
        self._load_local_api_keys()
        return
    
    def profile_exists(self, profile):
        return profile in self.profiles

    def change_profile(self, name):
        if name in self.profiles:
            self.profile = name

    def is_logged(self):
        if self.secret_api_key is not None:
            return True
        return False
    
    def fetch_secret_api_token(self, username:str, password:str):
        res = ops.fetch_secret_api_token(username, password)
        if res.get('status_code')!=200:
            return False, None
        self._update_local_api_keys(username, username, res.get('secret_api_key'), res.get('secret_api_key_uuid'))
        self._load_local_api_keys()
        self.profile = username
        api_key_uuid = res.get('secret_api_key_uuid')
        return True, api_key_uuid

    @property
    def secret_api_key(self):
        if self.profile in self.profiles:
            if 'secret_api_key' in self.profiles[self.profile]:
                return self.profiles[self.profile]["secret_api_key"]
        else:
            return None

    def _load_local_api_keys(self):
        self.profiles = {}
        if os.path.isfile(self.tokens_path):
            with open(self.tokens_path, 'r') as f:
                self.profiles = toml.load(f)
            if len(self.profiles)==1:
                self.profile = list(self.profiles.keys())[0]                
        return

    def _update_local_api_keys(self, profile, username, secret_api_key, secret_api_key_uuid):
        self._load_local_api_keys()
        self.profiles[profile] = {'secret_api_key': secret_api_key, 
                                    'username': username, 
                                    'api_key_uuid': secret_api_key_uuid}
        if not os.path.exists(self.tokens_folder):
            os.mkdir(self.tokens_folder)

        with open(self.tokens_path, 'w') as f:
            toml.dump(self.profiles, f)

    @property
    def tokens_path(self):
        home = str(Path.home())
        config_folder = os.path.join(home, '.hectiqlab')
        return os.path.join(config_folder, 'credentials')
    
    @property
    def tokens_folder(self):
        home = str(Path.home())
        config_folder = os.path.join(home, '.hectiqlab')
        return config_folder



import tensorflow as tf
from pyhectiqlab import Run

class LabCallback(tf.keras.callbacks.Callback):

    def __init__(self, run: Run, postfix: str):
        self._run = run
        self.epoch = 0
        self.global_step = 0
        self.postfix = postfix

    def on_train_begin(self, logs=None):
        self._run.training()

    def on_train_end(self, logs=None):
        self._run.completed()

    def on_epoch_begin(self, epoch, logs=None):
        pass

    def on_epoch_end(self, epoch, logs=None):
        self.epoch += 1

    def on_train_batch_end(self, batch, logs=None):
        self.global_step += 1
        for metric in logs.keys():
            metric_val = logs[metric]
            self._run.add_metrics(f'{metric}{self.postfix}', value=metric_val, step=self.global_step)

    def on_train_batch_begin(self, batch, logs=None):
        pass


import unittest
import uuid

import numpy as np
from pyhectiqlab import Run, Config, AuthProvider

run_name = f'test {uuid.uuid4()}'
run = Run(name=run_name, project='test', should_create_project=True)

class CreateRunTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.hectiq_run = run
        logger = cls.hectiq_run.start_recording_logs('unittest')
        
    def test_is_logged(self):
        self.assertEqual(AuthProvider().is_logged(), True)

    def test_add_meta(self):
        self.hectiq_run.add_meta('is_test', True)
        self.hectiq_run.add_meta('scalar', 1.2)
        self.hectiq_run.add_meta('string', 'Hello world')

    def test_add_tag(self):
        self.hectiq_run.add_tag(name='hello', description='This run is a test')
        self.hectiq_run.add_tag(name='world')

    def test_add_packages(self):
        self.hectiq_run.add_package_versions(globals())

    def test_add_package_repo_state(self):
        self.hectiq_run.add_package_repo_state('pyhectiqlab')

    def test_add_metrics(self):
        key = 'sinx'
        self.hectiq_run.set_metrics_cache_settings(min_cache_flush_delay=0, max_cache_length=10)
        for step in np.arange(100):
            value = np.sin(step/20)
            self.hectiq_run.add_metrics(key, value, step)

    def test_add_artifact(self):
        self.hectiq_run.add_artifacts('./test_img1.jpeg')
        self.hectiq_run.add_artifacts('./test_img2.jpeg')

    def test_add_step_artifact(self):
        self.hectiq_run.add_artifacts('./test_img2.jpeg', step=1)

    def test_add_config(self):
        sub_config = Config(c='hello world', d=1.5, e=True)
        config = Config(a=1, sub_config=sub_config)
        self.hectiq_run.add_config(config=config)

    def test_log(self):
        logger = self.hectiq_run.start_recording_logs()
        logger.info('info')
        logger.warning('warning')
        logger.error('error')
        logger.critical('critical')
        logger.debug('debug')

    def test_status(self):
        self.hectiq_run.running()
        self.hectiq_run.pending()
        self.hectiq_run.training()
        self.hectiq_run.completed()

if __name__ == '__main__':
    suite = unittest.defaultTestLoader.loadTestsFromTestCase(CreateRunTest)
    result = unittest.TextTestRunner().run(suite)

    if result.wasSuccessful():
        run.completed()
        run.add_tag(name='success')
    else:
        run.failed()


import threading

class RepeatedTimer(threading.Thread):
    def __init__(self, method, delay):
        super().__init__()
        self.daemon = True
        self._stop = threading.Event()
        self.delay = delay
        self.method = method

    def run(self):
        while not self._stop.wait(self.delay):
            self.method()
            
    def stopped(self):
        return self._stop.isSet()

    def stop(self):
        self._stop.set()


import os

if os.environ.get("HECTIQLAB_ENV")=="dev":
	server_url = "http://0.0.0.0:8080"
	app_url = "http://0.0.0.0:3000"
else:
	server_url = "https://hectiq-lab-api-rz3jqlxf4a-ue.a.run.app"
	app_url = "https://lab.hectiq.ai"